{-# OPTIONS_GHC -Wall -XFlexibleInstances #-}
{-
- Module to generate invariants from a parsed emod structure
- For information on how it works, please refer to the article
-  
-
- Copyright 2013 Sebastiaan J.C. Joosten
- Licensed under the Apache License, Version 2.0 (the "License");
-  you may not use this file except in compliance with the License.
-  You may obtain a copy of the License at
-
-      http://www.apache.org/licenses/LICENSE-2.0
-
-  Unless required by applicable law or agreed to in writing, software
-  distributed under the License is distributed on an "AS IS" BASIS,
-  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-  See the License for the specific language governing permissions and
-  limitations under the License.
-}
module ToINV (showINV)  where
import Terms (Term(..))
import qualified Data.Set as Set
import qualified Data.Map.Lazy as Map
import EParser (QueueAssignment(..),FlopAssignment(..),InputAssignment(..),OutputAssignment(..)
               ,OutputWire(..))
import Prelude hiding (init)
type LPMap a = Map.Map a Int
type UntimedLPTerm = Set.Set String

-- some functions for modifying linear equations expressed as data maps
(.*.) :: Int -> Map.Map a Int -> Map.Map a Int 
(.*.) x y = Map.map ((*) x) y
(.-.),(.+.) :: Ord a => Map.Map a Int -> Map.Map a Int -> Map.Map a Int
(.-.) = Map.mergeWithKey (\_ a b->case a-b of {0->Nothing;s->Just s}) id (Map.map (\x -> -x))
(.+.) = Map.mergeWithKey (\_ a b->case a+b of {0->Nothing;s->Just s}) id id

-- these are the primitive functions we can translate
data TranslatableTerm
 = TR_OR    Term Term
 | TR_XOR   Term Term
 | TR_ITE   Term Term Term
 | TR_NOT   Term
 | TR_VALUE Bool
 | TR_RESET deriving Show

-- "translateTranslatableTerm" as in Equations (1)
translateTranslatableTerm :: [TranslatableTerm]
                          -> LPMap UntimedLPTerm -- the resulting LP variable
translateTranslatableTerm [] = (Map.singleton Set.empty 1)
translateTranslatableTerm (ini:bs)
 = combineAnd (case ini of
                (TR_RESET    ) -> (Map.empty)
                (TR_VALUE b  ) -> if b then (Map.singleton Set.empty 1)
                                       else (Map.empty)
                (TR_ITE a b c) -> translateTerm (T_AND a b) .+. translateTerm (T_AND (T_NOT a) c) -- note: this is more efficient than translating T_ITE to a T_OR: we know that both sides of the T_OR are mutually exclusive and use that here
                (TR_OR  y1 y2) -> let t1=translateTerm ( y1)
                                      t2=translateTerm ( y2)
                                  in t1 .+. t2 .-. combineAnd t1 t2
                (TR_XOR y1 y2) -> let t1=translateTerm y1
                                      t2=translateTerm y2
                                  in t1 .+. t2 .-. (2 .*. combineAnd t1 t2)
                (TR_NOT y1   ) -> (Map.singleton Set.empty 1) .-. translateTerm y1
              )
              (translateTranslatableTerm bs)

-- "combineAnd" is some sort of Cartesian product, written as a circle with a cross in it (\otimes) in the article 
combineAnd :: LPMap UntimedLPTerm -> LPMap UntimedLPTerm -- two LPs for which the AND must be calculated
           -> LPMap UntimedLPTerm -- result
combineAnd a b = Map.filter ((/=) 0) $ 
                 Map.fromListWith (+) [(Set.union a' b', ai*bi) | (a',ai) <- Map.toList a, (b',bi) <- Map.toList b]

-- compute the LP-term for (x && y) in which x is a list of primitives and y is an LP-term
extendLPMap :: Set.Set String -> LPMap UntimedLPTerm -> LPMap UntimedLPTerm
extendLPMap s = Map.mapKeysWith (+) (Set.union s)

translateTerm :: Term -> LPMap UntimedLPTerm
translateTerm t = extendLPMap (Set.fromList s) (translateTranslatableTerm tt)
 where (s,tt) = collectAnds t
       collectAnds :: Term -> ([String],[TranslatableTerm])
       -- collectAnds:
       -- returns Either:
       --   Left: a set of strings which is to be used for declarations
       --   Right: tuple: ( list of conjuncts that are primitive and succesfully translated as a set of strings to be used for declarations
       --                 , list of translatables (i.e. not T_AND or primitive)
       --                   which may still contain T_AND in a nested place (such as T_NOT (T_AND ...))
       --                 )
       collectAnds (T_QTRDY x  ) = trm  "QTRDY" x
       collectAnds (T_QIRDY x  ) = trm  "QIRDY" x
       collectAnds (T_QDATA x y) = trm' "QDATA" x y
       collectAnds (T_OTRDY x  ) = trm  "OTRDY" x
       collectAnds (T_IIRDY x  ) = trm  "IIRDY" x
       collectAnds (T_IDATA x y) = trm' "IDATA" x y
       collectAnds (T_FLOPV x  ) = trm  "FLOPV" x
       collectAnds (T_UNKNOWN x) = trm  "UNKNO" (show x)
       collectAnds (T_INPUT x  ) = trm  "INPUT" x
       collectAnds (T_AND x y)
         = case (collectAnds x,collectAnds y) of
             ~((nx,ox),(ny,oy)) -> (nx++ny, ox++oy)
       collectAnds (T_OR  x y)   = wrong (TR_OR    x y)
       collectAnds (T_XOR x y)   = wrong (TR_XOR   x y)
       collectAnds (T_ITE x y z) = wrong (TR_ITE   x y z)
       collectAnds (T_NOT x)     = wrong (TR_NOT   x)
       collectAnds (T_VALUE x)   = wrong (TR_VALUE x)
       collectAnds (T_RESET)     = wrong (TR_RESET)
       trm x y = ([x++y],[])
       trm' x y u = ([x++y++"-"++show u],[])
       wrong x = ([],[x])

getPrimitives :: Ord a => [Map.Map a a1] -> Set.Set a
getPrimitives lst = Set.fromList (getPrimitives' lst)
getPrimitives' :: Ord a => [Map.Map a a1] -> [a]
getPrimitives' lst = concat (map Map.keys lst)

obtainDataWires :: QueueAssignment -> [(String, (QueueAssignment, Int))]
obtainDataWires o@(QueueAssignment qname _size _inIrdy idata _ouTrdy)
 = [(head (Set.toList (head (Map.keys (translateTerm (T_QDATA qname nr))))), (o,nr)) | nr<-[0..(length idata - 1)]]

showINV :: ([QueueAssignment],[FlopAssignment],[InputAssignment],[OutputAssignment],[OutputWire]) -> IO ()
showINV (queues,_flops,_ins,_outs,_wires)
 = putStrLn$ genEQS sol
 where preQueueInfo = obtainQueues queues
       primitives = Set.toList (getPrimitives (map snd queueInfo))
       dataWires = Map.fromList$ concat (map obtainDataWires queues)
       queueInfo = preQueueInfo ++ fixedPointExtraQueue dataWires (getPrimitives (map snd preQueueInfo))
       nrOfPrims = length primitives
       sol = fix [[fromIntegral (Map.findWithDefault 0 p qi) | p<-primitives] ++ take i (repeat 0) ++ 1:(take (length queueInfo - i - 1) (repeat 0)) | (i,(_,qi))<- zip [0..] queueInfo]
       fix (a:rst) = if (null trailing) then fix rst else a:fix (map sweep rst)
         where (leading,trailing) = span (\x -> x == 0) a
               -- trailNormed = map ((*) ht) trailing
               ht = head trailing
               ll = length leading
               sweep v = if f' == 0 then v else 
                         if f' == ht then [v' - a' | (a',v') <- zip a v] -- even though normalization might be possible, it is not (yet) required here. This reduces runtime with about 25% in practice
                         else normalize [v' * ht - a' * f' | (a',v') <- zip a v]
                 where f' = v !! ll
       fix [] = []
       normalize sp = [quot f factor | f <- sp]
        where factor = foldr1 gcd sp
       genEQS :: [[Int]] -> String
       genEQS [] = []
       genEQS (a:lst) = if and (map (\x -> x == 0) fp) then (negs ++" = "++poss)++"\n"++rst else rst
         where fp = take nrOfPrims a
               sp = drop nrOfPrims a
               poss = showAdd [( (v),j) | (v,j)<-zip sp [0..], v> 0]
               negs = showAdd [(-(v),j) | (v,j)<-zip sp [0..], v< 0]
               rst = genEQS lst
               showT (v,n) = if (v == 1) then nm else (show v)++"*"++nm
                 where nm = (fst (queueInfo !! n))
               showAdd [] = "0" 
               showAdd [e] = showT e
               showAdd (e:bs) = showT e ++" + "++showAdd bs

queueName :: QueueAssignment -> String
queueName (QueueAssignment qname _ _ _ _) = qname

fixedPointExtraQueue :: Map.Map String (QueueAssignment, Int) -> Set.Set (Set.Set String) -> [(String, Map.Map UntimedLPTerm Int)]
fixedPointExtraQueue datas prims
 = nub' Set.empty (getPrimSet (Set.toList prims))
 where
  getPrimSet ps
   = concat 
     [ Map.toList $
           Map.fromListWith (\(a,x) (_,y) -> (a,x++y)) 
                            [ (queueName q,(q,[nr]))
                            | var <- Set.toList prim
                            , Just (q, nr) <- [Map.lookup var datas]
                            ]
          | prim <- ps]
  nub' _ [] = []
  nub' prev ((qnm,(q,nrs)):rst)
   = if (Set.member (qnm,nrs) prev) then nub' prev rst else
     (qnm++"_"++show nrs, res):nub' (Set.insert (qnm,nrs) prev)
      (rst ++ [ps | ps@(a,(_,c)) <- getPrimSet (Map.keys res), not (Set.member (a,c) prev)])
   where (res1,res2) = (obtainExtraQueue q nrs)
         res = res1 .-. res2

obtainExtraQueue :: QueueAssignment -> [Int] -- list of data wire numbers
                    -> (Map.Map UntimedLPTerm Int,Map.Map UntimedLPTerm Int)
obtainExtraQueue (QueueAssignment qname _size inIrdy idata ouTrdy) vs
 = (enter , exit)
 where enter   = foldr combineAnd (translateTerm (T_AND inIrdy (T_QTRDY qname)))
                                  [translateTerm (idata !! v) | v<-vs] -- note: we could be much faster by not translating this as often as we do now
       exit    = foldr combineAnd (translateTerm (T_AND ouTrdy (T_QIRDY qname)))
                                  [translateTerm (T_QDATA qname v) | v<-vs]

obtainQueues :: [QueueAssignment] -> [(String, Map.Map UntimedLPTerm Int)]
obtainQueues [] = []
obtainQueues (QueueAssignment nm _size inIrdy _idata ouTrdy : lst)
 = (nm, enter .-. exit):obtainQueues lst
 where enter   = translateTerm (T_AND inIrdy (T_QTRDY nm))
       exit    = translateTerm (T_AND ouTrdy (T_QIRDY nm))
